A collection of data types, classes, and functions for primitve marshaling

Designed by the FFI task force,
hacked by Sven Panne <Sven.Panne@informatik.uni-muenchen.de>, Jan 2000

\begin{code}
#include "MachDeps.h"

module FFI
        ( -- {Int,Word}{8,16,32,64} are abstract and instance of Eq, Ord,
          -- Num, Bounded, Real, Integral, Ix, Enum, Read, Show
          -- NOTE 1: GHC has additional instances CCallable, CReturnable, Bits
          -- NOTE 2: No plain Word
          Int8,  Int16,  Int32,  Int64
        , Word8, Word16, Word32, Word64

        , Addr              -- abstract, instance of: Eq, Ord, Enum, Show
        , nullAddr          -- :: Addr
        , plusAddr          -- :: Addr -> Int -> Addr
			    
        , ForeignObj        -- abstract, instance of: Eq
        , makeForeignObj    -- :: Addr -> IO () -> IO ForeignObj
        , foreignObjToAddr  -- :: ForeignObj -> Addr

        , StablePtr,        -- abstract
        , makeStablePtr     -- :: a -> IO (StablePtr a)
        , deRefStablePtr    -- :: StablePtr a -> IO a
        , freeStablePtr     -- :: StablePtr a -> IO ()
        , stablePtrToAddr   -- :: StablePtr a -> Addr
        , addrToStablePtr   -- :: Addr -> StablePtr a

        , Storable(
             sizeOf,        -- :: a -> Int
             alignment,     -- :: a -> Int
             peekElemOff,   -- :: Addr -> Int      -> IO a
             pokeElemOff,   -- :: Addr -> Int -> a -> IO ()
             peekByteOff,   -- :: Addr -> Int      -> IO a
             pokeByteOff,   -- :: Addr -> Int -> a -> IO ()
             peek,          -- :: Addr             -> IO a
             poke)          -- :: Addr        -> a -> IO ()
			    
        , malloc,           -- ::                    Int -> IO Addr
        , mallocElem,       -- :: Storable a => a        -> IO Addr
        , mallocElems,      -- :: Storable a => a -> Int -> IO Addr
        , realloc,          -- ::            Addr -> Int -> IO Addr
        , free              -- :: Addr -> IO ()

        , isBigEndian       -- :: Bool
        ) where
\end{code}

\begin{code}
import Int        ( Int8,  Int16,  Int32,  Int64
                  , readInt8OffAddr,  writeInt8OffAddr
                  , readInt16OffAddr, writeInt16OffAddr
                  , readInt32OffAddr, writeInt32OffAddr
                  , readInt64OffAddr, writeInt64OffAddr
                  )
import Word       ( Word8, Word16, Word32, Word64
                  , readWord8OffAddr,  writeWord8OffAddr
                  , readWord16OffAddr, writeWord16OffAddr
                  , readWord32OffAddr, writeWord32OffAddr
                  , readWord64OffAddr, writeWord64OffAddr
                  )
import Addr       ( Addr, nullAddr, plusAddr
                  , readCharOffAddr,   writeCharOffAddr
                  , readIntOffAddr,    writeIntOffAddr
                  , readAddrOffAddr,   writeAddrOffAddr
                  , readFloatOffAddr,  writeFloatOffAddr
                  , readDoubleOffAddr, writeDoubleOffAddr
                  )
import ForeignObj ( ForeignObj, makeForeignObj, foreignObjToAddr )
import StablePtr  ( StablePtr, makeStablePtr, deRefStablePtr
                  ,  freeStablePtr, stablePtrToAddr, addrToStablePtr
                  )
import IOExts     ( fixIO )
import PrelAddr   ( Word, intToWord )
\end{code}

Primitive marshaling

Minimal complete definition: sizeOf, alignment, and one definition
in each of the peek/poke families.

\begin{code}
class Storable a where

   -- sizeOf/alignment *never* use their first argument
   sizeOf      :: a -> Int
   alignment   :: a -> Int

   -- replacement for read-/write???OffAddr
   peekElemOff :: Addr -> Int      -> IO a
   pokeElemOff :: Addr -> Int -> a -> IO ()

   -- the same with *byte* offsets
   peekByteOff :: Addr -> Int      -> IO a
   pokeByteOff :: Addr -> Int -> a -> IO ()

   -- ... and with no offsets at all
   peek        :: Addr             -> IO a
   poke        :: Addr        -> a -> IO ()

   -- circular default instances
   peekElemOff addr off = fixIO (\val -> peekByteOff addr (off * sizeOf val))
   pokeElemOff addr off val = pokeByteOff addr (off * sizeOf val) val

   peekByteOff addr off = peek (addr `plusAddr` off)
   pokeByteOff addr off = poke (addr `plusAddr` off)

   peek addr = peekElemOff addr 0
   poke addr = pokeElemOff addr 0
\end{code}

System-dependent, but rather obvious instances

\begin{code}
instance Storable Char where
   sizeOf      = const SIZEOF_CHAR
   alignment   = const ALIGNMENT_CHAR
   peekElemOff = readCharOffAddr
   pokeElemOff = writeCharOffAddr

instance Storable Int where
   sizeOf      = const SIZEOF_INT
   alignment   = const ALIGNMENT_INT
   peekElemOff = readIntOffAddr
   pokeElemOff = writeIntOffAddr

instance Storable Addr where
   sizeOf      = const SIZEOF_VOID_P
   alignment   = const ALIGNMENT_VOID_P
   peekElemOff = readAddrOffAddr
   pokeElemOff = writeAddrOffAddr

instance Storable Float where
   sizeOf      = const SIZEOF_FLOAT
   alignment   = const ALIGNMENT_FLOAT
   peekElemOff = readFloatOffAddr
   pokeElemOff = writeFloatOffAddr

instance Storable Double where
   sizeOf      = const SIZEOF_DOUBLE
   alignment   = const ALIGNMENT_DOUBLE
   peekElemOff = readDoubleOffAddr
   pokeElemOff = writeDoubleOffAddr

instance Storable Word8 where
   sizeOf      = const SIZEOF_WORD8
   alignment   = const ALIGNMENT_WORD8
   peekElemOff = readWord8OffAddr
   pokeElemOff = writeWord8OffAddr

instance Storable Word16 where
   sizeOf      = const SIZEOF_WORD16
   alignment   = const ALIGNMENT_WORD16
   peekElemOff = readWord16OffAddr
   pokeElemOff = writeWord16OffAddr

instance Storable Word32 where
   sizeOf      = const SIZEOF_WORD32
   alignment   = const ALIGNMENT_WORD32
   peekElemOff = readWord32OffAddr
   pokeElemOff = writeWord32OffAddr

instance Storable Word64 where
   sizeOf      = const SIZEOF_WORD64
   alignment   = const ALIGNMENT_WORD64
   peekElemOff = readWord64OffAddr
   pokeElemOff = writeWord64OffAddr

instance Storable Int8 where
   sizeOf      = const SIZEOF_INT8
   alignment   = const ALIGNMENT_INT8
   peekElemOff = readInt8OffAddr
   pokeElemOff = writeInt8OffAddr

instance Storable Int16 where
   sizeOf       = const SIZEOF_INT16
   alignment    = const ALIGNMENT_INT16
   peekElemOff  = readInt16OffAddr
   pokeElemOff  = writeInt16OffAddr

instance Storable Int32 where
   sizeOf      = const SIZEOF_INT32
   alignment   = const ALIGNMENT_INT32
   peekElemOff = readInt32OffAddr
   pokeElemOff = writeInt32OffAddr

instance Storable Int64 where
   sizeOf      = const SIZEOF_INT64
   alignment   = const ALIGNMENT_INT64
   peekElemOff = readInt64OffAddr
   pokeElemOff = writeInt64OffAddr
\end{code}

(de-)allocation of raw bytes

\begin{code}
malloc :: Int -> IO Addr
malloc unused = failWhenNULL "malloc" (malloc_ (intToWord unused))

mallocElem :: Storable a => a -> IO Addr
mallocElem unused = malloc (sizeOf unused)

mallocElems :: Storable a => a -> Int -> IO Addr
mallocElems unused numElems = malloc (numElems * sizeOf unused)

realloc :: Addr -> Int -> IO Addr
realloc oldAddr numBytes =
   failWhenNULL "realloc" (realloc_ oldAddr (intToWord numBytes))

foreign import ccall "free" unsafe free :: Addr -> IO ()
\end{code}

utility functions, not exported

\begin{code}
failWhenNULL :: String -> IO Addr -> IO Addr
failWhenNULL fun act = do
   addr <- act
   if addr == nullAddr
      then ioError (userError (fun ++ ": out of memory"))
      else return addr

-- TODO: Word should always match size_t
foreign import ccall "malloc"  unsafe malloc_  ::         Word -> IO Addr
foreign import ccall "realloc" unsafe realloc_ :: Addr -> Word -> IO Addr
\end{code}

Byte-ordering:

\begin{code}
isBigEndian :: Bool
isBigEndian = 
#ifdef WORDS_BIGENDIAN
    True
#else
    False
#endif
\end{code}
